import math

from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design_base.design import Design
from elasticai.creator.vhdl.design_base.ports import Port
from elasticai.creator.vhdl.designs.rom import Rom
from elasticai.creator.vhdl.savable import Path


class FPConv1d(Design):
    def __init__(
        self,
        name: str,
        total_bits: int,
        frac_bits: int,
        in_channels: int,
        out_channels: int,
        signal_length: int,
        kernel_size: int,
        weights: list[list[list[int]]],
        bias: list[int],
        stride: int,
        padding: int,
        dilation: int,
    ) -> None:
        super().__init__(name=name)
        self._total_bits = total_bits
        self._frac_bits = frac_bits
        self._in_channels = in_channels
        self._out_channels = out_channels
        self._input_signal_length = signal_length
        self._kernel_size = kernel_size
        self._weights = weights
        self._bias = bias
        self._stride = stride
        self._padding = padding
        self._output_signal_length = math.floor(
            (
                self._input_signal_length
                + 2 * self._padding
                - dilation * (self._kernel_size - 1)
                - 1
            )
            / self._stride
            + 1
        )

    @property
    def port(self) -> Port:
        return create_port(
            x_width=self._total_bits,
            y_width=self._total_bits,
            x_count=self._input_signal_length * self._in_channels,
            y_count=self._output_signal_length * self._out_channels,
        )

    @staticmethod
    def _flatten_params(params: list[list[list[int]]]) -> list[int]:
        result = []
        for list_of_lists in params:
            for list_of_int in list_of_lists:
                result.extend(list_of_int)
        return result

    def save_to(self, destination: Path) -> None:
        rom_name = dict(weights=f"{self.name}_w_rom", bias=f"{self.name}_b_rom")
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="fp_conv1d.tpl.vhd",
            parameters=dict(
                total_bits=str(self._total_bits),
                frac_bits=str(self._frac_bits),
                in_channels=str(self._in_channels),
                out_channels=str(self._out_channels),
                kernel_size=str(self._kernel_size),
                stride=str(self._stride),
                padding=str(self._padding),
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)

        weights_rom = Rom(
            name=rom_name["weights"],
            data_width=self._total_bits,
            values_as_integers=self._flatten_params(self._weights),
        )
        weights_rom.save_to(destination.create_subpath(rom_name["weights"]))

        bias_rom = Rom(
            name=rom_name["bias"],
            data_width=self._total_bits,
            values_as_integers=self._bias,
        )
        bias_rom.save_to(destination.create_subpath(rom_name["bias"]))



from typing import Optional, Protocol

import torch


class Arithmetics(Protocol):
    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def round(self, a: torch.Tensor) -> torch.Tensor:
        ...

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def sum(
        self, a: torch.Tensor, dim: Optional[int | tuple[int, ...]]
    ) -> torch.Tensor:
        ...

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        ...

    def conv1d(
        self,
        inputs: torch.Tensor,
        weights: torch.Tensor,
        bias: Optional[torch.Tensor],
        stride: int,
        padding: int | str,
        dilation: int,
        groups: int,
    ) -> torch.Tensor:
        ...



from typing import Optional, cast

import torch

from elasticai.creator.base_modules.arithmetics.arithmetics import Arithmetics
from elasticai.creator.base_modules.autograd_functions.round_to_float import (
    RoundToFloat,
)


class FloatArithmetics(Arithmetics):
    def __init__(self, mantissa_bits: int, exponent_bits: int) -> None:
        self.mantissa_bits = mantissa_bits
        self.exponent_bits = exponent_bits

    @property
    def largest_positive_value(self) -> float:
        exponent_bias = 2 ** (self.exponent_bits - 1)
        return (2 - 1 / 2**self.mantissa_bits) * 2 ** (
            2**self.exponent_bits - exponent_bias - 1
        )

    @property
    def smallest_negative_value(self) -> float:
        return -self.largest_positive_value

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(a))

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        return torch.clamp(
            a, min=self.smallest_negative_value, max=self.largest_positive_value
        )

    def round(self, a: torch.Tensor) -> torch.Tensor:
        return cast(
            torch.Tensor,
            RoundToFloat.apply(a, self.mantissa_bits, self.exponent_bits),
        )

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a + b)

    def sum(
        self, a: torch.Tensor, dim: Optional[int | tuple[int, ...]] = None
    ) -> torch.Tensor:
        return self.quantize(torch.sum(a, dim=dim))

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a * b)

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))

    def conv1d(
        self,
        inputs: torch.Tensor,
        weights: torch.Tensor,
        bias: Optional[torch.Tensor],
        stride: int,
        padding: int | str,
        dilation: int,
        groups: int,
    ) -> torch.Tensor:
        outputs = torch.nn.functional.conv1d(
            input=inputs,
            weight=weights,
            bias=bias,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
        )
        return self.quantize(outputs)



from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design_base.design import Design, Port
from elasticai.creator.vhdl.savable import Path


class FPHardSigmoid(Design):
    def __init__(
        self,
        name: str,
        total_bits: int,
        frac_bits: int,
        one: int,
        zero_threshold: int,
        one_threshold: int,
        slope: int,
        y_intercept: int,
    ) -> None:
        super().__init__(name)
        self._total_bits = total_bits
        self._frac_bits = frac_bits
        self._one = one
        self._zero_threshold = zero_threshold
        self._one_threshold = one_threshold
        self._slope = slope
        self._y_intercept = y_intercept

    @property
    def port(self) -> Port:
        return create_port(x_width=self._total_bits, y_width=self._total_bits)

    def save_to(self, destination: Path) -> None:
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="fp_hard_sigmoid.tpl.vhd",
            parameters=dict(
                layer_name=self.name,
                data_width=str(self._total_bits),
                frac_width=str(self._frac_bits),
                one=str(self._one),
                zero_threshold=str(self._zero_threshold),
                one_threshold=str(self._one_threshold),
                slope=str(self._slope),
                y_intercept=str(self._y_intercept),
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)



from typing import Any, cast

import torch

from elasticai.creator.base_modules.arithmetics.fixed_point_arithmetics import (
    FixedPointArithmetics,
)
from elasticai.creator.base_modules.linear import Linear
from elasticai.creator.base_modules.two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.nn.linear.design import FPLinear
from elasticai.creator.vhdl.translatable import Translatable


class FPBatchNormedLinear(Translatable, torch.nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool,
        total_bits: int,
        frac_bits: int,
        bn_eps: float = 1e-5,
        bn_momentum: float = 0.1,
        bn_affine: bool = True,
        device: Any = None,
    ) -> None:
        super().__init__()
        self._arithmetics = FixedPointArithmetics(
            config=FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
        )
        self._linear = Linear(
            in_features=in_features,
            out_features=out_features,
            arithmetics=self._arithmetics,
            bias=bias,
            device=device,
        )
        self._batch_norm = torch.nn.BatchNorm1d(
            num_features=out_features,
            eps=bn_eps,
            momentum=bn_momentum,
            affine=bn_affine,
            track_running_stats=True,
            device=device,
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        x = self._linear(inputs)
        x = self._batch_norm(x)
        return self._arithmetics.quantize(x)

    def translate(self, name: str) -> FPLinear:
        def float_to_signed_int(value: float | list) -> int | list:
            if isinstance(value, list):
                return list(map(float_to_signed_int, value))
            return self._arithmetics.config.as_integer(value)

        bn_mean = cast(torch.Tensor, self._batch_norm.running_mean)
        bn_variance = cast(torch.Tensor, self._batch_norm.running_var)
        bn_epsilon = self._batch_norm.eps
        lin_weight = self._linear.weight
        lin_bias = (
            torch.tensor([0] * self._linear.out_features)
            if self._linear.bias is None
            else self._linear.bias
        )

        std = torch.sqrt(bn_variance + bn_epsilon)
        weights = lin_weight / std
        bias = (lin_bias - bn_mean) / std

        if self._batch_norm.affine:
            weights = (self._batch_norm.weight * weights.t()).t()
            bias = self._batch_norm.weight * bias + self._batch_norm.bias

        return FPLinear(
            in_feature_num=self._linear.in_features,
            out_feature_num=self._linear.out_features,
            total_bits=self._arithmetics.config.total_bits,
            frac_bits=self._arithmetics.config.frac_bits,
            weights=cast(list[list[int]], float_to_signed_int(weights.tolist())),
            bias=cast(list[int], float_to_signed_int(bias.tolist())),
            name=name,
        )



from typing import Any

import torch

from .arithmetics.arithmetics import Arithmetics


class Conv1d(torch.nn.Conv1d):
    def __init__(
        self,
        arithmetics: Arithmetics,
        in_channels: int,
        out_channels: int,
        kernel_size: int | tuple[int],
        stride: int | tuple[int] = 1,
        padding: int | tuple[int] | str = 0,
        dilation: int | tuple[int] = 1,
        groups: int = 1,
        bias: bool = True,
        device: Any = None,
        dtype: Any = None,
    ) -> None:
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            padding_mode="zeros",
            device=device,
            dtype=dtype,
        )
        self._arithmetics = arithmetics

    @staticmethod
    def _flatten_tuple(x: int | tuple[int, ...]) -> int:
        return x[0] if isinstance(x, tuple) else x

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        quantized_weights = self._arithmetics.quantize(self.weight)
        quantized_bias = (
            self._arithmetics.quantize(self.bias) if self.bias is not None else None
        )
        return self._arithmetics.conv1d(
            inputs=inputs,
            weights=quantized_weights,
            bias=quantized_bias,
            stride=self._flatten_tuple(self.stride),
            padding=(
                self.padding
                if isinstance(self.padding, str)
                else self._flatten_tuple(self.padding)
            ),
            dilation=self._flatten_tuple(self.dilation),
            groups=self.groups,
        )



from typing import Optional

from elasticai.creator.vhdl.code_generation.template import Template, TemplateExpander
from elasticai.creator.vhdl.savable import File, Path


class InMemoryFile(File):
    def __init__(self, name: str) -> None:
        self.text: list[str] = []
        self.name = name

    def write(self, template: Template) -> None:
        expander = TemplateExpander(template)
        unfilled_variables = expander.unfilled_variables()
        if len(unfilled_variables) > 0:
            raise KeyError(
                "Template is not filled completly. The following variables are"
                f" unfilled: {', '.join(unfilled_variables)}."
            )
        for line in expander.lines():
            self.text.append(line)


class InMemoryPath(Path):
    def __init__(self, name: str, parent: Optional["InMemoryPath"]) -> None:
        self.name = name
        self.children: dict[str, InMemoryPath | InMemoryFile] = dict()
        self.parent = parent

    def as_file(self, suffix: str) -> InMemoryFile:
        file = InMemoryFile(f"{self.name}{suffix}")
        if len(self.children) > 0:
            raise ValueError(
                f"non empty path {self.name}, "
                f"present children: {', '.join(self.children)}"
            )
        if self.parent is not None:
            self.parent.children[self.name] = file
        return file

    def __getitem__(self, item: str) -> "InMemoryPath | InMemoryFile":
        return self.children[item]

    def create_subpath(self, subpath_name: str) -> "InMemoryPath":
        subpath = InMemoryPath(name=subpath_name, parent=self)
        self.children[subpath_name] = subpath
        return subpath



from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design_base.design import Design, Port
from elasticai.creator.vhdl.savable import Path


class FPHardTanh(Design):
    def __init__(
        self, name: str, total_bits: int, frac_bits: int, min_val: int, max_val: int
    ) -> None:
        super().__init__(name)
        self._total_bits = total_bits
        self._frac_bits = frac_bits
        self._min_val = min_val
        self._max_val = max_val

    @property
    def port(self) -> Port:
        return create_port(x_width=self._total_bits, y_width=self._total_bits)

    def save_to(self, destination: Path) -> None:
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="fp_hard_tanh.tpl.vhd",
            parameters=dict(
                layer_name=self.name,
                data_width=str(self._total_bits),
                frac_width=str(self._frac_bits),
                min_val=str(self._min_val),
                max_val=str(self._max_val),
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)



from .batch_normed_linear import FPBatchNormedLinear
from .conv1d import FPConv1d
from .hard_sigmoid import FPHardSigmoid
from .hard_tanh import FPHardTanh
from .identity import BufferedIdentity, BufferlessIdentity
from .linear import FPLinear
from .precomputed import FPSigmoid, FPTanh
from .relu import FPReLU
from .sequential import Sequential



import torch

from elasticai.creator.base_modules.arithmetics.arithmetics import Arithmetics


class SiLU(torch.nn.SiLU):
    def __init__(self, arithmetics: Arithmetics) -> None:
        super().__init__(inplace=False)
        self._arithmetics = arithmetics
        self.scale = torch.nn.Parameter(torch.ones(1, requires_grad=True))
        self.beta = torch.nn.Parameter(torch.zeros(1, requires_grad=True))

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        a = self.scale * super().forward(input) + self.beta
        x = self._arithmetics.quantize(a)
        return x



from elasticai.creator.base_modules.hard_tanh import HardTanh
from elasticai.creator.base_modules.two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.vhdl.design_base.design import Design
from elasticai.creator.vhdl.translatable import Translatable

from .design import FPHardTanh as FPHardTanhDesign


class FPHardTanh(Translatable, HardTanh):
    def __init__(
        self, total_bits: int, frac_bits: int, min_val: float = -1, max_val: float = 1
    ) -> None:
        super().__init__(min_val, max_val)
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)

    def translate(self, name: str) -> Design:
        return FPHardTanhDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            min_val=self._config.as_integer(self.min_val),
            max_val=self._config.as_integer(self.max_val),
        )



from elasticai.creator.base_modules.identity import Identity
from elasticai.creator.vhdl.design_base.design import Design
from elasticai.creator.vhdl.translatable import Translatable

from .design import BufferedIdentity as IdentityDesign
from .design import BufferlessDesign


class BufferedIdentity(Translatable, Identity):
    def __init__(self, num_input_features: int, total_bits: int) -> None:
        self._num_input_features = num_input_features
        self._num_input_bits = total_bits
        super().__init__()

    def translate(self, name: str) -> Design:
        return IdentityDesign(
            name=name,
            num_input_features=self._num_input_features,
            num_input_bits=self._num_input_bits,
        )


class BufferlessIdentity(Translatable, Identity):
    def __init__(self, total_bits: int) -> None:
        self._num_input_bits = total_bits
        super().__init__()

    def translate(self, name: str) -> Design:
        return BufferlessDesign(
            name=name,
            num_input_bits=self._num_input_bits,
        )



from typing import Optional, cast

import torch

from elasticai.creator.base_modules.autograd_functions.round_to_fixed_point import (
    RoundToFixedPoint,
)
from elasticai.creator.base_modules.two_complement_fixed_point_config import (
    FixedPointConfig,
)

from .arithmetics import Arithmetics


class FixedPointArithmetics(Arithmetics):
    def __init__(self, config: FixedPointConfig) -> None:
        self.config = config

    def quantize(self, a: torch.Tensor) -> torch.Tensor:
        return self.round(self.clamp(a))

    def clamp(self, a: torch.Tensor) -> torch.Tensor:
        return torch.clamp(
            a, min=self.config.minimum_as_rational, max=self.config.maximum_as_rational
        )

    def round(self, a: torch.Tensor) -> torch.Tensor:
        return cast(torch.Tensor, RoundToFixedPoint.apply(a, self.config))

    def add(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.clamp(a + b)

    def sum(
        self, a: torch.Tensor, dim: Optional[int | tuple[int, ...]] = None
    ) -> torch.Tensor:
        return self.clamp(torch.sum(a, dim=dim))

    def mul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(a * b)

    def matmul(self, a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
        return self.quantize(torch.matmul(a, b))

    def conv1d(
        self,
        inputs: torch.Tensor,
        weights: torch.Tensor,
        bias: torch.Tensor | None,
        stride: int,
        padding: int | str,
        dilation: int,
        groups: int,
    ) -> torch.Tensor:
        outputs = torch.nn.functional.conv1d(
            input=inputs,
            weight=weights,
            bias=bias,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
        )
        return self.quantize(outputs)



from typing import Any

import torch

from .arithmetics.arithmetics import Arithmetics


class Linear(torch.nn.Linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        arithmetics: Arithmetics,
        bias: bool,
        device: Any = None,
        dtype: Any = None,
    ) -> None:
        super().__init__(in_features, out_features, bias, device, dtype)
        self.ops = arithmetics

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        weight = self.ops.quantize(self.weight)

        if self.bias is not None:
            bias = self.ops.quantize(self.bias)
            return self.ops.add(self.ops.matmul(x, weight.T), bias)

        return self.ops.matmul(x, weight.T)



from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design_base.design import Design, Port
from elasticai.creator.vhdl.savable import Path


class FPReLU(Design):
    def __init__(self, name: str, data_width: int, use_clock: bool) -> None:
        super().__init__(name)
        self._data_width = data_width
        self._clock_option = "true" if use_clock else "false"

    @property
    def port(self) -> Port:
        return create_port(x_width=self._data_width, y_width=self._data_width)

    def save_to(self, destination: Path) -> None:
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="fp_relu.tpl.vhd",
            parameters=dict(
                layer_name=self.name,
                data_width=str(self._data_width),
                clock_option=self._clock_option,
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)



from elasticai.creator.base_modules.hard_sigmoid import HardSigmoid
from elasticai.creator.base_modules.two_complement_fixed_point_config import (
    FixedPointConfig,
)
from elasticai.creator.vhdl.design_base.design import Design
from elasticai.creator.vhdl.translatable import Translatable

from .design import FPHardSigmoid as FPHardSigmoidDesign


class FPHardSigmoid(Translatable, HardSigmoid):
    def __init__(self, total_bits: int, frac_bits: int) -> None:
        super().__init__()
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)

    def translate(self, name: str) -> Design:
        return FPHardSigmoidDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            one=self._config.as_integer(1),
            zero_threshold=self._config.as_integer(-3),
            one_threshold=self._config.as_integer(3),
            slope=self._config.as_integer(1 / 6),
            y_intercept=self._config.as_integer(0.5),
        )



from functools import partial, reduce
from itertools import chain

from elasticai.creator.vhdl.auto_wire_protocols.autowiring import (
    AutoWirer,
    DataFlowNode,
)
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.code_generation.code_abstractions import (
    create_connections_using_to_from_pairs,
    create_instance,
    create_signal_definitions,
)
from elasticai.creator.vhdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design_base.design import Design
from elasticai.creator.vhdl.design_base.ports import Port
from elasticai.creator.vhdl.savable import Path


class Sequential(Design):
    def __init__(
        self,
        sub_designs: list[Design],
        *,
        name: str,
    ) -> None:
        super().__init__(name)
        self._subdesigns = sub_designs
        self._connections: dict[tuple[str, str], tuple[str, str]] = (
            self._build_connections_map()
        )
        self._port = self._build_port()
        self._library_name_for_instances = "work"
        self._architecture_name_for_instances = "rtl"

    def _qualified_signal_name(self, instance: str, signal: str) -> str:
        return f"{instance}_{signal}"

    def _build_port(self) -> Port:
        def get_designs_by_name(name: str) -> list[Design]:
            return [d for d in self._subdesigns if d.name == name]

        def connected_to_self_source(
            sink_source: tuple[tuple[str, str], tuple[str, str]], self_source_name: str
        ) -> bool:
            return sink_source[1] == (self.name, self_source_name)

        def connected_to_self_sink(sink_source, self_sink_name):
            return sink_source[0] == (self.name, self_sink_name)

        def sink_design_name(sink_source):
            return sink_source[0][0]

        def source_design_name(
            sink_source: tuple[tuple[str, str], tuple[str, str]]
        ) -> str:
            return sink_source[1][0]

        def get_connected_designs(get_name, is_connected) -> list[Design]:
            return reduce(
                lambda a, b: a + b,
                map(
                    get_designs_by_name,
                    map(
                        get_name,
                        filter(is_connected, self._connections.items()),
                    ),
                ),
                [],
            )

        width = {k: 1 for k in ("x", "y_address", "x_address", "y")}
        sink_keys = ("y", "x_address")
        source_keys = ("x", "y_address")

        for key in sink_keys:
            connected = partial(connected_to_self_sink, self_sink_name=key)
            for d in get_connected_designs(source_design_name, connected):
                width[key] = d.port[key].width

        for key in source_keys:
            connected = partial(connected_to_self_source, self_source_name=key)
            for d in get_connected_designs(sink_design_name, connected):
                width[key] = d.port[key].width

        return create_port(
            x_width=width["x"],
            y_width=width["y"],
            x_address_width=width["x_address"],
            y_address_width=width["y_address"],
        )

    @property
    def port(self) -> Port:
        return self._port

    def _save_subdesigns(self, destination: Path) -> None:
        for design in self._subdesigns:
            design.save_to(destination.create_subpath(design.name))

    def _instance_names(self) -> list[str]:
        return [f"i_{design.name}" for design in self._subdesigns]

    def _build_connections_map(self) -> dict[tuple[str, str], tuple[str, str]]:
        named_ports = [(d.name, d.port) for d in self._subdesigns]
        nodes = [
            DataFlowNode(
                name=n[0],
                sinks=tuple(s.name for s in n[1].incoming),
                sources=tuple(s.name for s in n[1].outgoing),
            )
            for n in named_ports
        ]
        top = DataFlowNode.top(self.name)
        autowirer = AutoWirer()
        autowirer.wire(top, graph=nodes)
        return autowirer.connections()

    def _generate_connections_code(self) -> list[str]:
        def generate_name(node_name: str, signal_name: str) -> str:
            if node_name == self.name:
                return signal_name
            else:
                return "_".join(("i", node_name, signal_name))

        map = {
            generate_name(*k): generate_name(*v) for k, v in self._connections.items()
        }

        lines = create_connections_using_to_from_pairs(map)
        lines = list(sorted(lines))
        return lines

    def _instance_name_and_design_pairs(self):
        yield from zip(self._instance_names(), self._subdesigns)

    def _generate_instantiations(self) -> list[str]:
        instantiations: list[str] = list()
        for instance, design in self._instance_name_and_design_pairs():
            signal_map = {
                signal.name: self._qualified_signal_name(instance, signal.name)
                for signal in design.port
            }
            instantiations.extend(
                create_instance(
                    name=instance,
                    entity=design.name,
                    library=self._library_name_for_instances,
                    architecture=self._architecture_name_for_instances,
                    signal_mapping=signal_map,
                )
            )
        return instantiations

    def _generate_signal_definitions(self) -> list[str]:
        return sorted(
            chain.from_iterable(
                create_signal_definitions(f"{instance_id}_", instance.port.signals)
                for instance_id, instance in self._instance_name_and_design_pairs()
            )
        )

    @property
    def _x_address_width(self) -> int:
        return self.port["x_address"].width

    @property
    def _y_address_width(self) -> int:
        return self.port["y_address"].width

    @property
    def _x_width(self) -> int:
        return self.port["x"].width

    @property
    def _y_width(self) -> int:
        return self.port["y"].width

    def save_to(self, destination: Path):
        self._save_subdesigns(destination)
        network_template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="network.tpl.vhd",
            parameters=dict(
                layer_connections=self._generate_connections_code(),
                layer_instantiations=self._generate_instantiations(),
                signal_definitions=self._generate_signal_definitions(),
                x_address_width=str(self._x_address_width),
                y_address_width=str(self._y_address_width),
                x_width=str(self._x_width),
                y_width=str(self._y_width),
                layer_name=self.name,
            ),
        )
        destination.create_subpath(self.name).as_file(".vhd").write(network_template)



from elasticai.creator.base_modules.arithmetics.fixed_point_arithmetics import (
    FixedPointArithmetics,
)
from elasticai.creator.base_modules.silu import SiLU
from elasticai.creator.base_modules.two_complement_fixed_point_config import (
    FixedPointConfig,
)

from .fp_monotonic_increasing_module import FPPrecomputedMonotonicIncreasingModule


class FPSiLU(FPPrecomputedMonotonicIncreasingModule):
    def __init__(
        self,
        total_bits: int,
        frac_bits: int,
        num_steps: int,
        sampling_intervall: tuple[float, float] = (-10, 10),
    ) -> None:
        super().__init__(
            base_module=SiLU(
                arithmetics=FixedPointArithmetics(
                    config=FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)
                ),
            ),
            total_bits=total_bits,
            frac_bits=frac_bits,
            num_steps=num_steps,
            sampling_intervall=sampling_intervall,
        )



from abc import ABC, abstractmethod

from elasticai.creator.vhdl.design_base.ports import Port
from elasticai.creator.vhdl.savable import Path, Savable


class Design(Savable, ABC):
    def __init__(self, name: str):
        self.name = name

    @property
    @abstractmethod
    def port(self) -> Port:
        ...

    @abstractmethod
    def save_to(self, destination: Path) -> None:
        ...



from typing import Any

import torch

from elasticai.creator.base_modules.two_complement_fixed_point_config import (
    FixedPointConfig,
)


class RoundToFixedPoint(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> torch.Tensor:
        if len(args) != 2:
            raise TypeError(
                "apply() takes exactly two arguments "
                "(x: torch.Tensor, config: FixedPointConfig)"
            )
        x: torch.Tensor = args[0]
        config: FixedPointConfig = args[1]

        fp_ints = config.as_integer(x)
        out_of_bounds = fp_ints[config.integer_out_of_bounds(fp_ints)]
        if torch.any(out_of_bounds):
            raise ValueError("Cannot quantize tensor. Values out of bounds.")

        return config.as_rational(fp_ints)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None



from collections.abc import Callable

from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.code_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.design_base.design import Design
from elasticai.creator.vhdl.design_base.ports import Port
from elasticai.creator.vhdl.savable import Path


class PrecomputedMonotonicIncreasingScalarFunction(Design):
    _template_package = module_to_package(__name__)

    def __init__(
        self,
        name: str,
        width: int,
        function: Callable[[int], int],
        inputs: list[int],
    ) -> None:
        super().__init__(name)
        self._width = width
        self._function = function
        self._inputs = inputs
        self._template = InProjectTemplate(
            file_name="precomputed_monotonic_increasing_scalar_function.tpl.vhd",
            package=self._template_package,
            parameters=dict(name=self.name, data_width=str(width)),
        )

    def _compute_io_pairs(self) -> dict[int, int]:
        inputs_in_descending_order = sorted(self._inputs, reverse=True)
        pairs = dict()
        for number in inputs_in_descending_order:
            pairs[number] = self._function(number)
        return pairs

    @property
    def port(self) -> Port:
        return create_port(x_width=self._width, y_width=self._width)

    def save_to(self, destination: Path) -> None:
        process_content = []

        pairs = list(self._compute_io_pairs().items())
        input_value, output_value = pairs[0]
        process_content.append(
            f"if signed_x <= {input_value} then "
            f"signed_y <= to_signed({output_value}, {self._width});"
        )
        for input_value, output_value in pairs[1:-1]:
            process_content.append(
                f"elsif signed_x <= {input_value} then "
                f"signed_y <= to_signed({output_value}, {self._width});"
            )
        _, output = pairs[-1]
        process_content.append(f"else signed_y <= to_signed({output}, {self._width});")
        process_content.append("end if;")

        self._template.parameters.update(process_content=process_content)
        destination.create_subpath(self.name).as_file(".vhd").write(self._template)



from typing import Any

import torch


class IdentityStepFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> torch.Tensor:
        if len(args) != 2:
            raise TypeError(
                "apply() takes exactly two arguments "
                "(inputs: torch.Tensor, step_lut: torch.Tensor)"
            )

        inputs: torch.Tensor = args[0]
        step_lut: torch.Tensor = args[1]

        steps = len(step_lut)
        if steps < 2:
            raise ValueError(
                f"Number of steps cannot be less than or equal to 1 (steps == {steps})."
            )
        inputs = inputs.cpu().to(torch.float32)
        inputs = inputs.clamp(min=step_lut.min(), max=step_lut.max())

        for step_idx in range(1, len(step_lut)):
            prev_step, curr_step = step_lut[step_idx - 1], step_lut[step_idx]
            inputs[(inputs > prev_step) & (inputs <= curr_step)] = curr_step

        return inputs

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Any:
        return *grad_outputs, None
